from typing import *

import numpy as np
import pandas as pd


def cross_correlate(df: pd.DataFrame, relate_to_series: pd.Series, lag_idx=0):
    """Calculate cross correlation for a given lag.

    It is recommended to either have a lot of data in the data frame, or to use a short time frame for the lags,
    as the results are unstable if too few data points overlap in the time shifted time series.

    Args:
        df (pandas.Series): Time series data to correlate with some series
        relate_to_series (pandas.Series): Pandas Series with time series data to relate df to. Must have the same
            temporal spacing as df.
        lag_idx (int): How many indices to move the DataFrame in relation to the series.

    Returns:
        pandas.DataFrame: Pandas DataFrame containing the cross correlations of the columns.

    Examples:
        >>> df = pd.DataFrame({'x': (1, 7, 3, 5), 'y': (3, 7, 6, 4)})
        >>> cross_correlate(df, df['x'], lag_idx=1)
        x   -0.981981
        y   -0.960769
    """

    correlations = df.corrwith(relate_to_series.shift(-lag_idx))
    return correlations


def columns_by_max_cross_correlation(
    df: pd.DataFrame, relate_to: Union[int, str], lags: pd.TimedeltaIndex, return_cross_correlation_df: bool = False
) -> Union[pd.DataFrame, Tuple[pd.DataFrame, pd.DataFrame]]:
    """Find lag of highest correlation and return relevant information for all columns of the inputted DataFrame.
    Note that the operation requires a DataFrame with even temporal spacing.

    It is recommended to either have a lot of data in the data frame, or to use a short time frame for the lags,
    as the results are unstable if too few data points overlap in the time shifted time series.

    Args:
        df (pandas.DataFrame): Time series data
        relate_to (Union[int, str]): Name of column to compare others with
        lags (pandas.TimedeltaIndex): Pandas sequence of timedeltas for shifting the time series
        return_cross_correlation_df (bool): Whether or not to return the cross correlations for all columns.
            This is a pandas.DataFrame containing the cross correlation for all calculated lags.
    Returns:
        Union[pandas.DataFrame, Tuple[pandas.DataFrame, pandas.DataFrame]]: Pandas DataFrame containing results of
        calculations, and optionally a DataFrame containing the cross correlations for each column at all calculated
        lags.

    Examples:
        Return maximum correlations and time lags for a simple dataframe.

        >>> df = pd.DataFrame({'datetime': pd.date_range(datetime(2017, 1, 1), datetime(2017, 1, 3), periods=10),
        >>>                    'x': np.sin(np.linspace(0, 2 * np.pi, 10)),
        >>>                    'y': np.sin(np.linspace(1, 2 * np.pi + 1, 10))}).set_index('datetime')
        >>> lags = pd.timedelta_range(timedelta(days=-3), timedelta(), periods=10)
        >>> columns_by_cross_correlation(df, 'x', lags)

    """
    diffs = df.index.to_series().diff()
    dmin, dmax = diffs.min(), diffs.max()
    assert dmin == dmax, "Time series must be evenly spaced"
    assert df.isna().sum().sum() == 0, "NaN values must be interpolated away before calculating cross-correlation"

    # Enforce most common time spacing in main column
    time_interval = dmin
    relate_to_df = df[relate_to]

    # Round lag timings to integer multiples of the minimum shifts
    lags_idx = np.unique(lags // time_interval).astype(int)

    # Interpolate for all levels of lag
    cross_correlations = np.zeros(shape=(lags_idx.shape[0], df.shape[1]))
    for i, l in enumerate(lags_idx):
        correlations = cross_correlate(df, relate_to_df, lag_idx=l)
        cross_correlations[i] = correlations

    # Find optimal lag for every feature
    max_corr_idxs = np.nanargmax(np.abs(cross_correlations), axis=0)
    max_lags = lags_idx[max_corr_idxs] * time_interval

    # Sorted indices by correlation
    max_corrs = cross_correlations[(max_corr_idxs, np.arange(max_corr_idxs.shape[0]))]
    sorted_idxs = np.abs(max_corrs).argsort()[::-1]

    out_df = pd.DataFrame(
        {"col": df.columns[sorted_idxs], "corr": max_corrs[sorted_idxs], "lag": -max_lags[sorted_idxs]}
    )
    if return_cross_correlation_df:
        cross_correlations = pd.DataFrame(cross_correlations)
        cross_correlations.set_index(lags_idx * time_interval, inplace=True)
        cross_correlations.columns = df.columns
        cross_correlations = cross_correlations[df.columns[sorted_idxs]]
        return out_df, cross_correlations
    return out_df



import datetime
from typing import *

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import ticker


def _time_ticks(x, pos=None):
    td = datetime.timedelta(seconds=int(abs(x)))
    if x < 0:
        return "– " + str(td)
    return str(td)


def _td_to_sec(td: datetime.timedelta):
    return td.days * 24 * 60 * 60 + td.seconds


def plot_cross_correlations(
    cross_correlation_df: pd.DataFrame,
    cols_to_plot: Union[Iterable[str], Iterable[int]] = None,
    separate_plots: bool = True,
    mpl_args: Dict[str, Any] = None,
) -> None:
    """Plot cross-correlations over time lags. The cols_to_plot parameter can either be an iterable of strings of
    columns to plot, or a list of integers of the indices of the sorted list of columns to display.

    Args:
        cross_correlation_df (pd.DataFrame): The DataFrame returned from columns_by_max_correlation (with the optional
            parameter return_cross_correlation_df enabled).
        cols_to_plot (Union[Iterable[str], Iterable[int]]): What columns to plot or, alternatively an iterable of
            indices for which of the ordered columns to plot.
        separate_plots (bool): Whether or not to divide the plots into individual plots, or to keep them in one plot
        mpl_args (Dict[str, Any]): Additional parameters for Matplotlib plotting function
    Returns:
        None

    Examples:
        Plot the cross-correlation for columns numbered 1-4 of the dataframe (0 is excluded to not plot auto-correlation).

        >>> plot_cross_correlation(cross_correlation_df, range(1,5))
    """
    if mpl_args is None:
        mpl_args = {}
    if cols_to_plot is not None:
        # Make indexable
        cols_to_plot = np.array(cols_to_plot)
        assert cols_to_plot.dtype.kind in ("i", "U"), "cols_to_plot must be an iterable of strings or ints"
        if cols_to_plot.dtype.kind == "i":
            assert np.all(
                (cols_to_plot < cross_correlation_df.shape[1]) & (cols_to_plot >= 0)
            ), "When indexing columns, the indices must be within range"
            cols_to_plot = cross_correlation_df.columns[cols_to_plot]
        assert np.all(
            [col in cross_correlation_df.columns for col in cols_to_plot]
        ), "All columns to plot must exist in the DataFrame"
        cross_correlation_df = cross_correlation_df[cols_to_plot]
    df_sec = cross_correlation_df.set_index(cross_correlation_df.index.to_series().apply(_td_to_sec))
    if separate_plots:
        for colname in df_sec.columns:
            # Information about plotted cross-correlation
            col = df_sec[colname]
            argmax = np.nanargmax(np.abs(col.values))
            maxval, maxtime = col.values[argmax], col.index[argmax]
            plt.figure(figsize=(15, 7))
            plt.plot(col, **mpl_args)
            plt.gca().axvline(maxtime, color="r")
            plt.suptitle(colname + "\n")
            plt.title("Max Correlation: {}, time lag: {}".format(str(maxval)[:6], str(_time_ticks(maxtime))))
            plt.gcf().set_facecolor("white")
            formatter = ticker.FuncFormatter(_time_ticks)
            plt.gca().xaxis.set_major_formatter(formatter)
            plt.xticks(rotation=60)
            plt.show(block=False)
    else:
        df_sec.plot(figsize=(20, 10), **mpl_args)
        plt.gcf().set_facecolor("white")
        formatter = ticker.FuncFormatter(_time_ticks)
        plt.gca().xaxis.set_major_formatter(formatter)
        plt.xticks(rotation=60)
        plt.show(block=False)



import re
from os import path

from setuptools import find_packages, setup

version = re.search('^__version__\s*=\s*"(.*)"', open("cognite/correlation/__init__.py").read(), re.M).group(1)

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="cognite-correlation",
    version=version,
    packages=["cognite." + p for p in find_packages(where="cognite")],
    install_requires=["pandas", "numpy", "matplotlib"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.5",
)
